#include <sys/_stdint.h>
#include <zephyr/kernel.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/drivers/gpio.h>

#include "sds011.h"

extern struct k_fifo fifo_adv_data;

#define UPDATE_MIN(MIN, SAMPLE) if (SAMPLE < MIN) { MIN = SAMPLE; }
#define UPDATE_MAX(MAX, SAMPLE) if (SAMPLE > MAX) { MAX = SAMPLE; }
#define UPDATE_MIN_MAX(MIN, MAX, SAMPLE) { UPDATE_MIN(MIN, SAMPLE); UPDATE_MAX(MAX, SAMPLE); }

#define UART_PORT DT_NODELABEL(uart0)
    static const struct device *const uart_dev = DEVICE_DT_GET(UART_PORT);

    uint8_t SDS_SLEEPCMD[19] = {
        0xAA, 0xB4, 0x06, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	0x00, 0x00, 0x00, 0xFF,	0xFF, 0x05, 0xAB
    };
    uint8_t SDS_WAKEUPCMD[19] = {
        0xAA, 0xB4, 0x06, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x06, 0xAB
    };
static K_FIFO_DEFINE(fifo_pms);
    struct pms_data_t {
        void *fifo_reserved;
        uint32_t ts;
        char data[10];
    };
    //char pm_measure_data[10];
    struct pms_data_t pms_data;
    int pm_measure_position = 0;
    bool pms_enable = false;
    bool pms_warming = false;
#define WARMUPTIME_SDS_MS 15000
#define READINGTIME_SDS_MS 5000

#define GPIO_PORT DT_NODELABEL(gpio0)
#define GPIO_PIN 22
const struct device *devgpio = DEVICE_DT_GET(GPIO_PORT);

uint16_t sds_pm10;
uint16_t sds_pm25;

bool uart_data_received = false;


uint32_t getTsSDS()
    {
        uint32_t ts = (uint32_t) (k_uptime_get()/1000);
        // LOG_INF("getTs: %d", ts);
        return ts;
    }

    void send_uart_cmd(uint8_t *buf)
    {
        //int msg_len = strlen(buf);
        //printk("Send to uart\n");

        for (int i = 0; i < 19; i++) {
            uart_poll_out(uart_dev, buf[i]);
        }
    }
void serial_cb(const struct device *dev, void *user_data)
    {
        uint8_t c;

        if (!uart_irq_update(uart_dev)) {
            return;
        }

        if (!uart_irq_rx_ready(uart_dev)) {
            return;
        }

        /* read until FIFO empty */
        while (uart_fifo_read(uart_dev, &c, 1) == 1) {
            if(c == 0xaa) { uart_data_received = true; } 
            if(pms_enable == true && pms_warming == false)
            {
                //printk("UART char received: 0x%x\n", c);
                if(c == 0xaa) { pm_measure_position = 0; pms_data.ts = getTsSDS(); }
                if(pm_measure_position >= 0) {
                    pms_data.data[pm_measure_position] = c;
                    pm_measure_position = pm_measure_position + 1;
                }
                if(pm_measure_position == 10) {
                    size_t size = sizeof(struct pms_data_t);
                    char *mem_ptr = k_malloc(size);
                    memcpy(mem_ptr, &pms_data, size);
                    k_fifo_put(&fifo_pms, mem_ptr);
                }
                //} else if(pms_enable == false) {
                //		send_uart_cmd(SDS_SLEEPCMD);
        } 
        }
    }

int sds011_setup()
{
      if (!device_is_ready(uart_dev)) {
            printk("UART device not found!");
            return -1;
        }

        const struct uart_config uartcfg = { 9600, UART_CFG_PARITY_NONE, UART_CFG_STOP_BITS_1, UART_CFG_DATA_BITS_8, UART_CFG_FLOW_CTRL_NONE };

        int ret = uart_configure(uart_dev, &uartcfg);
        if (ret < 0) {
                printk("Error in comfiguring UART\n");
        }

        ret = uart_irq_callback_user_data_set(uart_dev, serial_cb, NULL);

        if (ret < 0) {
            if (ret == -ENOTSUP) {
                printk("Interrupt-driven UART API support not enabled\n");
            } else if (ret == -ENOSYS) {
                printk("UART device does not support interrupt-driven API\n");
            } else {
                printk("Error setting UART callback: %d\n", ret);
            }
            return -1;
        }
        uart_irq_rx_enable(uart_dev);

        gpio_pin_configure(devgpio, GPIO_PIN, GPIO_OUTPUT | GPIO_OPEN_DRAIN);

        gpio_pin_set(devgpio, GPIO_PIN, 0);
        //send_uart_cmd(SDS_WAKEUPCMD);
        k_msleep(5000);
        if(uart_data_received == true) {
          return 0;
        } else {
          gpio_pin_set(devgpio, GPIO_PIN, 1);
          return -1;
        }
}



int sds011_probe()
    {
        uint32_t ts_start_measure;
        gpio_pin_set(devgpio, GPIO_PIN, 0);
        //send_uart_cmd(SDS_WAKEUPCMD);
        pms_warming = true;
        pms_enable = true;
        k_msleep(WARMUPTIME_SDS_MS);
        pms_warming = false;
        ts_start_measure = getTsSDS();

        struct pms_data_t *pms_data_read;

        uint32_t sds_pm10_sum = 0;
        uint32_t sds_pm25_sum = 0;
        uint32_t sds_val_count = 0;
        uint32_t sds_pm10_max = 0;
        uint32_t sds_pm10_min = 20000;
        uint32_t sds_pm25_max = 0;
        uint32_t sds_pm25_min = 20000;

        for(;;)
        {
            pms_data_read = k_fifo_get(&fifo_pms, K_MSEC(READINGTIME_SDS_MS));
            if(pms_data_read != NULL)
            {
                if(getTsSDS() < (ts_start_measure + (READINGTIME_SDS_MS/1000)))
                {
                    //printk("Parse PMS %d %d\n", ts_start_measure, getTsSDS());
                    uint32_t pm10_serial = 0;
                    uint32_t pm25_serial = 0;
                    int checksum_is = pms_data_read->data[2] + pms_data_read->data[3] + pms_data_read->data[4] + pms_data_read->data[5] + pms_data_read->data[6] + pms_data_read->data[7];
                    if(pms_data_read->data[8] == (checksum_is % 256) && pms_data_read->data[0] == 170 && pms_data_read->data[1] == 192 && pms_data_read->data[9] == 171)
                {

                        pm25_serial = pms_data_read->data[2] | pms_data_read->data[3] << 8;
                        pm10_serial = pms_data_read->data[4] | pms_data_read->data[5] << 8;
                        sds_pm10_sum += pm10_serial;
                        sds_pm25_sum += pm25_serial;
                        UPDATE_MIN_MAX(sds_pm10_min, sds_pm10_max, pm10_serial)
                        UPDATE_MIN_MAX(sds_pm25_min, sds_pm25_max, pm25_serial)

                            sds_val_count += 1;


                        printk("p10: %d p25: %d\n", pm10_serial, pm25_serial);
                    }
                } else {
                    send_uart_cmd(SDS_SLEEPCMD);
                    pms_enable = false;
                }
                k_free(pms_data_read);

            } else {
                break;
            }	
        }	

        gpio_pin_set(devgpio, GPIO_PIN, 1);

        if (sds_val_count > 2) {
            sds_pm10_sum = sds_pm10_sum - sds_pm10_min - sds_pm10_max;
            sds_pm25_sum = sds_pm25_sum - sds_pm25_min - sds_pm25_max;
            sds_val_count = sds_val_count - 2;
        }

        if (sds_val_count > 0) {
    sds_pm25 = sds_pm25_sum / (sds_val_count * 10);
    sds_pm10 = sds_pm10_sum / (sds_val_count * 10);

         /*   struct bthome_adv_t bthome_adv;
            bthome_adv.type = PM25;
            bthome_adv.data_u16 = sds_pm25_sum / (sds_val_count * 10);
            size_t size = sizeof(struct bthome_adv_t);
            char *mem_ptr = k_malloc(size);
            memcpy(mem_ptr, &bthome_adv, size);
            k_fifo_put(&fifo_adv_data, mem_ptr);
            printk("PM25: %d\n",bthome_adv.data_u16);
            bthome_adv.type = PM10;
            bthome_adv.data_u16 = sds_pm10_sum / (sds_val_count * 10);
            char *mem_ptr2 = k_malloc(size);
            memcpy(mem_ptr2, &bthome_adv, size);
            k_fifo_put(&fifo_adv_data, mem_ptr2);
            printk("PM10: %d\n",bthome_adv.data_u16);
    */
        }
  return 0;
    }

uint16_t sds011_ReadPM25() 
{
  return sds_pm25;
}

uint16_t sds011_ReadPM10() 
{
  return sds_pm10;
}
